
#include "hle.h"

HLE (memset)
{
	memset (HLE_PARAM_1_PTR, HLE_PARAM_2, HLE_PARAM_3);
	HLE_RETURN (HLE_PARAM_1);
}


HLE (memcpy)
{
	memcpy (HLE_PARAM_1_PTR, HLE_PARAM_2_PTR, HLE_PARAM_3);
	HLE_RETURN (HLE_PARAM_1);
}


HLE (memmove)
{
	memmove (HLE_PARAM_1_PTR, HLE_PARAM_2_PTR, HLE_PARAM_3);
	HLE_RETURN (HLE_PARAM_1);
}


HLE (bzero)
{
	bzero (HLE_PARAM_1_PTR, HLE_PARAM_2);
}


HLE (strlen)
{
	HLE_RETURN (strlen (HLE_PARAM_1_PTR));
}


HLE (strncpy)
{
	strncpy (HLE_PARAM_1_PTR, HLE_PARAM_2_PTR, HLE_PARAM_3);
	HLE_RETURN (HLE_PARAM_1);
}


HLE (strcpy)
{
	strcpy (HLE_PARAM_1_PTR, HLE_PARAM_2_PTR);
	HLE_RETURN (HLE_PARAM_1);
}


typedef struct
{
	__u16 buttons;
	__s8 stickX, stickY;
	__s8 substickX, substickY;
	__u8 trigL, trigR;
	__u8 analogA, analogB;
	__s8 err;
} PADStatus;

#define PAD_ERR_NONE						0
#define PAD_ERR_NO_CONTROLLER		-1
#define PAD_ERR_NOT_READY				-2

#include "hw_si.h"
HLE (PADRead)
{
	PADStatus *status = HLE_PARAM_1_PTR;
	int i;


	for (i = 0; i < 4; i++)
	{
		if (SI_CONTROLLER_TYPE (i) == SIDEV_GC_CONTROLLER)
		{
			status[i].buttons = BSWAP16 (RSI32 (SI_CHN_INH (i)) >> 16);
			status[i].stickX = 0x80 + (__s8) (RSI32 (SI_CHN_INH (i)) >> 8);
			status[i].stickY = 0x80 + (__u8) RSI32 (SI_CHN_INH (i));
/*
			printf ("PADRead: %.4x, (%d, %d), (%d, %d)\n",
							status[i].buttons, status[i].stickX, status[i].stickY, status[i].substickX, status[i].substickY);
*/
			status[i].err = PAD_ERR_NONE;
			RSI32 (SI_CHN_INH (i)) &= ~0x00800000;
		}
		else
			bzero (&status[i], sizeof (PADStatus));

		if (0 == SI_CONTROLLER_TYPE (i))
			status[i].err = PAD_ERR_NO_CONTROLLER;
	}

	// is motor present
	HLE_RETURN (FALSE);
}


HLE (PADInit)
{
}


HLE (PADReset)
{
	HLE_RETURN (TRUE);
}


HLE (SIProbe)
{
	HLE_RETURN (SI_CONTROLLER_TYPE (HLE_PARAM_1));
}


const char *movie_ext[] =
{
	"thp", "h4m", "str",
};


int is_movie_file (const char *name)
{
	int len = strlen (name);
	int i;


	if (len > 4)
		for (i = 0; i < sizeof (movie_ext) / sizeof (*movie_ext); i++)
			if (0 == strcmp (movie_ext[i], &name[len - 3]))
				return TRUE;
	
	return FALSE;
}


HLE (DVDOpen_ignore_movies)
{
	if (is_movie_file (HLE_PARAM_1_PTR))
		HLE_RETURN (0);

	HLE_EXECUTE_LLE;
}


HLE (DVDConvertPathToEntrynum_ignore_movies)
{
	if (is_movie_file (HLE_PARAM_1_PTR))
		HLE_RETURN (0);

	HLE_EXECUTE_LLE;
}
