/*
 *  gcube
 *  Copyright (c) 2004 monk
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 *
 *  DSP - digital signal processor interface (0xcc005000)
 *  AI  - audio interface (0xcc006c00)    
 *         
 */


#include "hw_ai_dsp.h"


__u8 rdsp[RDSP_SIZE];
__u8 rai[RAI_SIZE];

__u8 ARAM[ARAM_SIZE];

static int feed_dsp = FALSE;


__u16 dsp_r16_direct (__u32 addr)
{
	DEBUG (EVENT_LOG_DSP, ".dsp: read  [%.4x] (%.4x)", addr & 0xffff, RDSP16 (addr));

	return RDSP16 (addr);
}


void dsp_w16_direct (__u32 addr, __u16 data)
{
	DEBUG (EVENT_LOG_DSP, ".dsp: write [%.4x] (%.4x) = %.4x", addr & 0xffff, RDSP16 (addr), data);
	RDSP16 (addr) = data;
}


__u32 dsp_r32_direct (__u32 addr)
{
	DEBUG (EVENT_LOG_DSP, ".dsp: read  [%.4x] (%.8x)", addr & 0xffff, RDSP32 (addr));

	return RDSP32 (addr);
}


void dsp_w32_direct (__u32 addr, __u32 data)
{
	DEBUG (EVENT_LOG_DSP, ".dsp: write [%.4x] (%.8x) = %.8x", addr & 0xffff, RDSP32 (addr), data);
	RDSP32 (addr) = data;
}


__u32 ai_r32_direct (__u32 addr)
{
	DEBUG (EVENT_LOG_AI, "..ai: read  [%.4x] (%.8x)", addr & 0xffff, RAI32 (addr));
	return RAI32 (addr);
}


void ai_w32_direct (__u32 addr, __u32 data)
{
	DEBUG (EVENT_LOG_AI, "..ai: write [%.4x] (%.8x) = %.8x", addr & 0xffff, RAI32 (addr), data);
	RAI32 (addr) = data;
}


void ai_w32_cr (__u32 addr, __u32 data)
{
	DEBUG (EVENT_LOG_AI, "..ai: AICR | PSTAT %d | FREQ %d kHz | AIINT %d/%d | AIINTVLD %d | SCRESET %d",
				 data & AICR_PSTAT, ((data >> 1) & 1) ? 48 : 32, (data & AICR_AIINTMSK)>0, (data & AICR_AIINT)>0, (data & AICR_AIINTVLD)>0, (data & AICR_SCRESET)>0);

	audio_set_freq ((data & AICR_32KHZ) ? FREQ_32KHZ : FREQ_48KHZ);

	if (data & AICR_SCRESET)
	{
		AISCNT = 0;
		data &= ~AICR_SCRESET;
	}

	RAI32 (addr) = data;
}


// sample counter
__u32 ai_r32_aiscnt (__u32 addr)
{
	DEBUG (EVENT_LOG_AI, "..ai: AISCNT read %.8x", RAI32 (addr));

	// fake it
	return RAI32 (addr)++;
}


void ai_w32_aivr (__u32 addr, __u32 data)
{
	DEBUG (EVENT_LOG_AI, "..ai: VOLUME %d / %d", data & 0x00ff, data & 0xff00);
	RAI32 (addr) = data;
}


void dsp_w16_csr (__u32 addr, __u16 data)
{
	DEBUG (EVENT_LOG_DSP, ".dsp: (%.8x) CSR = %.4x %s %s", PC, data,
				 (data & DSP_CSR_RES) ? "reset" : "", (data & DSP_CSR_HALT) ? "halt" : "");

	if (data & DSP_CSR_PIINT)
		pi_interrupt (INTERRUPT_DSP);

	if (data & DSP_CSR_DSPINT)
		RDSP16 (DSP_CSR) &= ~DSP_CSR_DSPINT;

	if (data & DSP_CSR_ARINT)
		RDSP16 (DSP_CSR) &= ~DSP_CSR_ARINT;

	if (data & DSP_CSR_AIDINT)
		RDSP16 (DSP_CSR) &= ~DSP_CSR_AIDINT;

	if (data & DSP_CSR_DMAINT)
		RDSP16 (DSP_CSR) &= ~DSP_CSR_DMAINT;

	RDSP16 (DSP_CSR) |= data & (DSP_CSR_DSPINTMSK | DSP_CSR_ARINTMSK | DSP_CSR_AIDINTMSK);

	if (feed_dsp && (data & DSP_CSR_RES))
		RDSP32 (DSP_CPU_MBOX_H) = 0x8071feed;
}


__u32 mail_pending[4] = {0};

__u16 dsp_r16_mailbox (__u32 addr)
{
	static int counter = 0;


	if (counter++ == 2)
	{
		if (mail_pending[0])
		{
			int i;

			RDSP32 (addr) = mail_pending[0];
			for (i = 0; i < 3; i++)
				mail_pending[i] = mail_pending[i + 1];
			counter = 2;
		}
		else
		{
			RDSP16 (addr) ^= 0x8000;
			counter = 0;
		}
	}

	return (RDSP16 (addr));
}


// just a hack
void dsp_w16_mailbox (__u32 addr, __u16 data)
{
	static int initialize_dsp = FALSE;

	DEBUG (EVENT_LOG_DSP, ".dsp: MAIL [%.2x] %.4x", addr & 0xff, data);

	if (!(addr & 0xff) && !data)
		feed_dsp = TRUE;

	if (addr & 2)
	{
		if (initialize_dsp)
		{
			RDSP32 (DSP_CPU_MBOX_H) = 0xdcd10000;
			initialize_dsp = FALSE;
			
			RDSP16 (DSP_CSR) |= DSP_CSR_DSPINT;
			pi_interrupt (INTERRUPT_DSP);
		}

		// dsp mailbox low
		switch (data)
		{
			case 0xa001:
				DEBUG (EVENT_LOG_DSP, "      RAM_MMEM_ADDR");
				break;

			case 0xc002:
				DEBUG (EVENT_LOG_DSP, "      RAM_ADDR");
				break;

			case 0xa002:
				DEBUG (EVENT_LOG_DSP, "      RAM_LENGTH");
				break;

			case 0xb002:
				DEBUG (EVENT_LOG_DSP, "      ARAM_MMEM_ADDR");
				break;

			case 0xd001:
				DEBUG (EVENT_LOG_DSP, "      DSP_INIT_VECTOR");
	
				initialize_dsp = TRUE;			
				break;
			
			case 0x0040:
				RDSP32 (DSP_CPU_MBOX_H) = 0xdcd10004;
				// ? 0xf3550000 | DSP_MBOX_H
				mail_pending[0] = 0xdcd10004;
				// will get xored with 0x80000000
				mail_pending[1] = 0x73558100;

				RDSP16 (DSP_CSR) |= DSP_CSR_DSPINT;
				pi_interrupt (INTERRUPT_DSP);
				break;
		}
	}

	RDSP16 (0x5000) &= ~0x8000;
}


void dsp_w16_dmacr (__u32 addr, __u16 data)
{
	RDSP16 (addr) = data;

	if (data & DMACR_PLAY)
	{
		DEBUG (EVENT_LOG_DSP, ".dsp: DMACR start %.8x length %d",
					 RDSP32 (DSP_DMA_START_H) | 0x80000000,
					 (RDSP16 (DSP_DMACR) & 0x7fff) << 5);
#if ENABLE_SOUND
		if (data & 0x7fff)
		audio_play (MEM_ADDRESS (RDSP32 (DSP_DMA_START_H)),
								(data & 0x7fff) << 5);
#endif
	}
	else
	{
#if ENABLE_SOUND
		audio_stop ();
#endif
		DEBUG (EVENT_LOG_DSP, ".dsp: DMACR stop, length %d", (data & 0x7fff) << 5);
	}
}


// todo: aram must be configured properly.
void dsp_w32_arcnt (__u32 addr, __u32 data)
{
	RDSP32 (addr) = data &~ ARCNT_READ;

	if (data & ARCNT_READ)
	{
		// transfer from aram
		DEBUG (EVENT_LOG_DSP, ".dsp: read %.8x MEM[%.8x] <- ARAM[%.8x]",
					 RDSP32 (ARDMA_CNT_H), RDSP32 (ARDMA_MMADDR_H) | 0x80000000, RDSP32 (ARDMA_ARADDR_H));

		if (RDSP32 (ARDMA_ARADDR_H) + RDSP32 (ARDMA_CNT_H) <= ARAM_SIZE)
			memcpy (MEM_ADDRESS (RDSP32 (ARDMA_MMADDR_H)),
							ARAM_ADDRESS (RDSP32 (ARDMA_ARADDR_H) & ARAM_MASK),
							RDSP32 (ARDMA_CNT_H));
	}
	else
	{
		// transfer to aram
		DEBUG (EVENT_LOG_DSP, ".dsp: write %.8x MEM[%.8x] -> ARAM[%.8x]",
					 RDSP32 (ARDMA_CNT_H), RDSP32 (ARDMA_MMADDR_H) | 0x80000000, RDSP32 (ARDMA_ARADDR_H));

		if (RDSP32 (ARDMA_ARADDR_H) + RDSP32 (ARDMA_CNT_H) <= ARAM_SIZE)
			memcpy (ARAM_ADDRESS (RDSP32 (ARDMA_ARADDR_H) & ARAM_MASK),
							MEM_ADDRESS (RDSP32 (ARDMA_MMADDR_H)),
							RDSP32 (ARDMA_CNT_H));
	}

	RDSP32 (ARDMA_MMADDR_H) = 0;
	RDSP32 (ARDMA_ARADDR_H) = 0;
	RDSP32 (ARDMA_CNT_H) = 0;

	RDSP16 (DSP_CSR) |= DSP_CSR_ARINT;
	if (RDSP16 (DSP_CSR) & DSP_CSR_ARINTMSK)
		pi_interrupt (INTERRUPT_DSP);
}


void dsp_w16_arcntl (__u32 addr, __u16 data)
{
	RDSP16 (addr) = data;

	dsp_w32_arcnt (ARDMA_CNT_H, RDSP32 (ARDMA_CNT_H));
}


void dsp_finished (void)
{
	DEBUG (EVENT_LOG_DSP, ".dsp: finished");

	RDSP16 (DSP_CSR) |= DSP_CSR_AIDINT;
	if (RDSP16 (DSP_CSR) & DSP_CSR_AIDINTMSK)
		pi_interrupt_ex (INTERRUPT_DSP);

	RDSP16 (DSP_DMACR) &= ~DMACR_PLAY;
}


void dsp_reinit (void)
{
}


void dsp_init (void)
{
	memset (rdsp, 0, sizeof (rdsp));
	memset (rai, 0, sizeof (rai));

	RAI32 (AI_AICR) |= AICR_32KHZ;
	RDSP16 (DSP_CSR) |= DSP_CSR_HALT;
	RDSP32 (DSP_CPU_MBOX_H) = 0x80544348;

	// control status register
	mem_hwr_hook (16, DSP_CSR, dsp_r16_direct);
	mem_hww_hook (16, DSP_CSR, dsp_w16_csr);

	// dsp mailbox (to dsp)
	mem_hwr_hook (16, DSP_MBOX_H, dsp_r16_direct);
	mem_hww_hook (16, DSP_MBOX_H, dsp_w16_mailbox);
	mem_hwr_hook (16, DSP_MBOX_L, dsp_r16_direct);
	mem_hww_hook (16, DSP_MBOX_L, dsp_w16_mailbox);

	// cpu mailbox (from dsp)
	mem_hwr_hook (16, DSP_CPU_MBOX_H, dsp_r16_mailbox);
	mem_hww_hook (16, DSP_CPU_MBOX_H, dsp_w16_direct);
	mem_hwr_hook (16, DSP_CPU_MBOX_L, dsp_r16_direct);
	mem_hww_hook (16, DSP_CPU_MBOX_L, dsp_w16_direct);

	// ARAM mode
	RDSP16 (0x5016) = 1;
	mem_hwr_hook (16, 0x5016, dsp_r16_direct);
	
	// ARAM size
	RDSP16 (0x5012) = ARAM_SIZE >> 16;
//	mem_hww_hook (16, 0x5012, dsp_w16_direct);
	mem_hwr_hook (16, 0x5012, dsp_r16_direct);

	mem_hwr_hook (32, ARDMA_MMADDR_H, dsp_r32_direct);
	mem_hww_hook (32, ARDMA_MMADDR_H, dsp_w32_direct);
	mem_hwr_hook (16, ARDMA_MMADDR_H, dsp_r16_direct);
	mem_hww_hook (16, ARDMA_MMADDR_H, dsp_w16_direct);
	mem_hwr_hook (16, ARDMA_MMADDR_L, dsp_r16_direct);
	mem_hww_hook (16, ARDMA_MMADDR_L, dsp_w16_direct);

	mem_hwr_hook (32, ARDMA_ARADDR_H, dsp_r32_direct);
	mem_hww_hook (32, ARDMA_ARADDR_H, dsp_w32_direct);
	mem_hwr_hook (16, ARDMA_ARADDR_H, dsp_r16_direct);
	mem_hww_hook (16, ARDMA_ARADDR_H, dsp_w16_direct);
	mem_hwr_hook (16, ARDMA_ARADDR_L, dsp_r16_direct);
	mem_hww_hook (16, ARDMA_ARADDR_L, dsp_w16_direct);

	// ARDMA control / length
	mem_hww_hook (32, ARDMA_CNT_H, dsp_w32_arcnt);
	mem_hwr_hook (32, ARDMA_CNT_H, dsp_r32_direct);
	mem_hww_hook (16, ARDMA_CNT_H, dsp_w16_direct);
	mem_hww_hook (16, ARDMA_CNT_L, dsp_w16_arcntl);
	mem_hwr_hook (16, ARDMA_CNT_H, dsp_r16_direct);
	mem_hwr_hook (16, ARDMA_CNT_L, dsp_r16_direct);
	
	// DMA start address
	mem_hww_hook (16, DMA_START_H, dsp_w16_direct);
	mem_hwr_hook (16, DMA_START_H, dsp_r16_direct);
	mem_hww_hook (16, DMA_START_L, dsp_w16_direct);
	mem_hwr_hook (16, DMA_START_L, dsp_r16_direct);
	
	// DMA control / length
	mem_hww_hook (16, DMA_CR_LEN, dsp_w16_dmacr);
	mem_hwr_hook (16, DMA_CR_LEN, dsp_r16_direct);
	
	// DMA bytes left
	mem_hwr_hook (16, DMA_BLEFT,	dsp_r16_direct);

	// ai control register
	mem_hwr_hook (32, AI_AICR, ai_r32_direct);
	mem_hww_hook (32, AI_AICR, ai_w32_cr);
	
	// ai sample counter
	mem_hwr_hook (32, AI_AISCNT, ai_r32_aiscnt);

	// volume
	mem_hww_hook (32, AI_VR, ai_w32_aivr);
	mem_hwr_hook (32, AI_VR, ai_r32_direct);

	// ai interrupt timing
	mem_hww_hook (32, AI_IT, ai_w32_direct);
}
