#ifndef __HW_GX_H
#define __HW_GX_H 1


#include "general.h"
#include "mem.h"
#include "hw_cp.h"


//#define LOG_FAKE_SENDS							1

#ifdef LOG_FAKE_SENDS
# define LOG_FAKE										DEBUG
#else
# define LOG_FAKE(X,format,...)			({})
#endif

#define GPOPCODE(X)										static __s32 X (__u32 mem)

#define GX_QUADS											GL_QUADS
#define GX_TRIANGLES									GL_TRIANGLES
#define GX_TRIANGLE_STRIP							GL_TRIANGLE_STRIP
#define GX_TRIANGLE_FAN								GL_TRIANGLE_FAN
#define GX_LINES											GL_LINES
#define GX_LINE_STRIP									GL_LINE_STRIP
#define GX_POINTS											GL_POINTS

#define CP_VCD_LO											(CP (0x50))
#define CP_VCD_HI											(CP (0x60))
#define CP_VAT_A(X)										(CP (0x70 + X))
#define CP_VAT_B(X)										(CP (0x80 + X))
#define CP_VAT_C(X)										(CP (0x90 + X))
					
#define VCD_COL1											((CP_VCD_LO >> 15) & 3)
#define VCD_COL0											((CP_VCD_LO >> 13) & 3)
#define VCD_NORMAL										((CP_VCD_LO >> 11) & 3)
#define VCD_POS												((CP_VCD_LO >>  9) & 3)
#define VCD_MIDX											(CP_VCD_LO & 0x01ff)

#define VCD_TEX7											((CP_VCD_HI >> 14) & 3)
#define VCD_TEX6											((CP_VCD_HI >> 12) & 3)
#define VCD_TEX5											((CP_VCD_HI >> 10) & 3)
#define VCD_TEX4											((CP_VCD_HI >>  8) & 3)
#define VCD_TEX3											((CP_VCD_HI >>  6) & 3)
#define VCD_TEX2											((CP_VCD_HI >>  4) & 3)
#define VCD_TEX1											((CP_VCD_HI >>  2) & 3)
#define VCD_TEX0											((CP_VCD_HI >>  0) & 3)

#define VCD_T7MIDX										(CP_VCD_LO & 0x100)
#define VCD_T6MIDX										(CP_VCD_LO & 0x080)
#define VCD_T5MIDX										(CP_VCD_LO & 0x040)
#define VCD_T4MIDX										(CP_VCD_LO & 0x020)
#define VCD_T3MIDX										(CP_VCD_LO & 0x010)
#define VCD_T2MIDX										(CP_VCD_LO & 0x008)
#define VCD_T1MIDX										(CP_VCD_LO & 0x004)
#define VCD_T0MIDX										(CP_VCD_LO & 0x002)
#define VCD_PMIDX											(CP_VCD_LO & 0x001)

#define VAT_BYTE_DEQUANT(X)						((CP_VAT_A(X) >> 30) & 0x01)
#define VAT_TEX0_SHFT(X)							((CP_VAT_A(X) >> 25) & 0x1f)
#define VAT_TEX0_FMT(X)								((CP_VAT_A(X) >> 22) & 0x07)
#define VAT_TEX0_CNT(X)								((CP_VAT_A(X) >> 21) & 0x01)
#define VAT_COL1_FMT(X)								((CP_VAT_A(X) >> 18) & 0x07)
#define VAT_COL1_CNT(X)								((CP_VAT_A(X) >> 17) & 0x01)
#define VAT_COL0_FMT(X)								((CP_VAT_A(X) >> 14) & 0x07)
#define VAT_COL0_CNT(X)								((CP_VAT_A(X) >> 13) & 0x01)
#define VAT_NORMAL_FMT(X)							((CP_VAT_A(X) >> 10) & 0x07)
#define VAT_NORMAL_CNT(X)							((CP_VAT_A(X) >>  9) & 0x01)
#define VAT_POS_SHFT(X)								((CP_VAT_A(X) >>  4) & 0x1f)
#define VAT_POS_FMT(X)								((CP_VAT_A(X) >>  1) & 0x07)
#define VAT_POS_CNT(X)								((CP_VAT_A(X) >>  0) & 0x01)

#define VAT_TEX4_FMT(X)								((CP_VAT_B(X) >> 28) & 0x07)
#define VAT_TEX4_CNT(X)								((CP_VAT_B(X) >> 27) & 0x01)
#define VAT_TEX3_SHFT(X)							((CP_VAT_B(X) >> 22) & 0x1f)
#define VAT_TEX3_FMT(X)								((CP_VAT_B(X) >> 19) & 0x07)
#define VAT_TEX3_CNT(X)								((CP_VAT_B(X) >> 18) & 0x01)
#define VAT_TEX2_SHFT(X)							((CP_VAT_B(X) >> 13) & 0x1f)
#define VAT_TEX2_FMT(X)								((CP_VAT_B(X) >> 10) & 0x07)
#define VAT_TEX2_CNT(X)								((CP_VAT_B(X) >>  9) & 0x01)
#define VAT_TEX1_SHFT(X)							((CP_VAT_B(X) >>  4) & 0x1f)
#define VAT_TEX1_FMT(X)								((CP_VAT_B(X) >>  1) & 0x07)
#define VAT_TEX1_CNT(X)								((CP_VAT_B(X) >>  0) & 0x01)

#define VAT_TEX7_SHFT(X)							((CP_VAT_C(X) >> 27) & 0x1f)
#define VAT_TEX7_FMT(X)								((CP_VAT_C(X) >> 24) & 0x07)
#define VAT_TEX7_CNT(X)								((CP_VAT_C(X) >> 23) & 0x01)
#define VAT_TEX6_SHFT(X)							((CP_VAT_C(X) >> 18) & 0x1f)
#define VAT_TEX6_FMT(X)								((CP_VAT_C(X) >> 15) & 0x07)
#define VAT_TEX6_CNT(X)								((CP_VAT_C(X) >> 14) & 0x01)
#define VAT_TEX5_SHFT(X)							((CP_VAT_C(X) >>  9) & 0x1f)
#define VAT_TEX5_FMT(X)								((CP_VAT_C(X) >>  6) & 0x07)
#define VAT_TEX5_CNT(X)								((CP_VAT_C(X) >>  5) & 0x01)
#define VAT_TEX4_SHFT(X)							((CP_VAT_C(X) >>  0) & 0x1f)

#define XF_PROJECTION_A								(FXF (0x1020))
#define XF_PROJECTION_B								(FXF (0x1021))
#define XF_PROJECTION_C								(FXF (0x1022))
#define XF_PROJECTION_D								(FXF (0x1023))
#define XF_PROJECTION_E								(FXF (0x1024))
#define XF_PROJECTION_F								(FXF (0x1025))
#define XF_PROJECTION_ORTHOGRAPHIC		(XF  (0x1026))

#define XF_VIEWPORT_A									(FXF (0x101a))
#define XF_VIEWPORT_B									(FXF (0x101b))
#define XF_VIEWPORT_C									(FXF (0x101c))
#define XF_VIEWPORT_D									(FXF (0x101d))
#define XF_VIEWPORT_E									(FXF (0x101e))
#define XF_VIEWPORT_F									(FXF (0x101f))

#define XF_INVTXSPEC									(XF  (0x1008))
#define HOST_TEXTURES									((XF_INVTXSPEC >> 4) & 15)
#define HOST_NORMALS									((XF_INVTXSPEC >> 2) &  3)
#define HOST_COLORS										((XF_INVTXSPEC >> 0) &  3)

#define XF_MATERIAL0									(XF (0x100c))
#define XF_MATERIAL1									(XF (0x100d))

#define XF_MATRIX_INDEX_0							(XF  (0x1018))
#define XF_MATRIX_INDEX_1							(XF  (0x1019))
#define MIDX_TEX7											((XF_MATRIX_INDEX_1 >> 18) & 0x3f)
#define MIDX_TEX6											((XF_MATRIX_INDEX_1 >> 12) & 0x3f)
#define MIDX_TEX5											((XF_MATRIX_INDEX_1 >>  6) & 0x3f)
#define MIDX_TEX4											((XF_MATRIX_INDEX_1 >>  0) & 0x3f)
#define MIDX_TEX3											((XF_MATRIX_INDEX_0 >> 24) & 0x3f)
#define MIDX_TEX2											((XF_MATRIX_INDEX_0 >> 18) & 0x3f)
#define MIDX_TEX1											((XF_MATRIX_INDEX_0 >> 12) & 0x3f)
#define MIDX_TEX0											((XF_MATRIX_INDEX_0 >>  6) & 0x3f)
#define MIDX_GEO											((XF_MATRIX_INDEX_0 >>  0) & 0x3f)

#define CP_VERTEX_ARRAY_BASE					(CP (0xa0))
#define CP_COLOR_ARRAY_BASE						(CP (0xa2))
#define CP_TEXTURE_ARRAY_BASE					(CP (0xa4))
#define CP_VERTEX_ARRAY_STRIDE				(CP (0xb0))
#define CP_COLOR_ARRAY_STRIDE					(CP (0xb2))
#define CP_TEXTURE_ARRAY_STRIDE				(CP (0xb4))

#define BP_LPSIZE											(BP (0x22))
#define LINE_ASPECT										((BP_LPSIZE >> 22) &    1)
#define POINT_TEX_OFFSET							((BP_LPSIZE >> 19) & 0x0f)
#define LINE_TEX_OFFSET								((BP_LPSIZE >> 16) & 0x0f)
#define POINT_SIZE										((BP_LPSIZE >>  8) & 0xff)
#define LINE_SIZE											((BP_LPSIZE >>  0) & 0xff)

#define BP_SSIZE(X)										(BP (0x30 + X*2 + 0))
#define BP_TSIZE(X)										(BP (0x30 + X*2 + 1))
#define TEX_SSIZE(X)									((BP_SSIZE(X) & 0xffff) + 1)
#define TEX_TSIZE(X)									((BP_TSIZE(X) & 0xffff) + 1)

#define BP_SETMODE0(X)								(BP (0x80 | (X & 3) | ((X & 4) << 3)))
#define BP_SETMODE1(X)								(BP (0x84 | (X & 3) | ((X & 4) << 3)))
#define BP_SETIMAGE0(X)								(BP (0x88 | (X & 3) | ((X & 4) << 3)))
#define BP_SETIMAGE1(X)								(BP (0x8c | (X & 3) | ((X & 4) << 3)))
#define BP_SETIMAGE2(X)								(BP (0x90 | (X & 3) | ((X & 4) << 3)))
#define BP_SETIMAGE3(X)								(BP (0x94 | (X & 3) | ((X & 4) << 3)))
#define BP_SETTLUT(X)									(BP (0x98 | (X & 3) | ((X & 4) << 3)))

#define BP_LOADTLUT0									(BP (0x64))
#define BP_LOADTLUT1									(BP (0x65))

#define BP_GENMODE										(BP (0x0))
#define BP_ZMODE											(BP (0x40))
#define BP_CMODE0											(BP (0x41))
#define BP_CMODE1											(BP (0x42))

#define BP_COPY_CLEAR_AR							(BP (0x4f))
#define BP_COPY_CLEAR_GB							(BP (0x50))
#define BP_COPY_CLEAR_Z								(BP (0x51))
#define BP_COPY_EXECUTE								(BP (0x52))

#define BP_SCISSORS_XY								(BP (0x20))
#define BP_SCISSORS_WH								(BP (0x21))

#define CULL_MODE											((BP_GENMODE >> 14) & 3)

#define ZMODE_UPDATE_ENABLED					(BP_ZMODE & 16)
#define ZMODE_FUNC										((BP_ZMODE >> 1) & 7)
#define ZMODE_ENABLED									(BP_ZMODE & 1)

#define CMODE_CONSTANT_ALPHA_ENABLED	(BP_CMODE1 & 256)
#define CMODE_CONSTANT_ALPHA					((BP_CMODE1 >>  0) & 0xff)
#define CMODE_LOGICOP									((BP_CMODE0 >> 12) & 0xf)
#define CMODE_SFACTOR									((BP_CMODE0 >>  8) & 7)
#define CMODE_DFACTOR									((BP_CMODE0 >>  5) & 7)
#define CMODE_ALPHA_ENABLED						(BP_CMODE0 & 16)
#define CMODE_COLOR_ENABLED						(BP_CMODE0 &  8)
#define CMODE_DITHER_ENABLED					(BP_CMODE0 &  4)
#define CMODE_BLEND										(BP_CMODE0 &  3)

#define COPY_CLEAR_A									((BP_COPY_CLEAR_AR >> 8) & 0xff)
#define COPY_CLEAR_R									((BP_COPY_CLEAR_AR >> 0) & 0xff)
#define COPY_CLEAR_G									((BP_COPY_CLEAR_GB >> 8) & 0xff)
#define COPY_CLEAR_B									((BP_COPY_CLEAR_GB >> 0) & 0xff)
#define COPY_CLEAR_Z									(BP_COPY_CLEAR_Z)

#define COPY_EXECUTE_TO_XFB						(BP_COPY_EXECUTE & 0x4000)
#define COPY_EXECUTE_CLEAR_EFB				(BP_COPY_EXECUTE & 0x0800)
#define COPY_EXECUTE_GAMMA						((BP_COPY_EXECUTE >> 7) & 3)

#define TEX_FORMAT(X)									(((BP_SETIMAGE0(X) >> 20) & 0x00f))
#define TEX_HEIGHT(X)									(((BP_SETIMAGE0(X) >> 10) & 0x3ff) + 1)
#define TEX_WIDTH(X)									(((BP_SETIMAGE0(X) >>  0) & 0x3ff) + 1)
#define TEX_IMAGE_BASE(X)							((BP_SETIMAGE3(X) & 0x00ffffff) << 5)
#define TEX_ODD_TMEM(X)								((BP_SETIMAGE2(X) & 0x0000ffff) << 5)
#define TEX_EVEN_TMEM(X)							((BP_SETIMAGE1(X) & 0x0000ffff) << 5)

#define TLUT_FORMAT(X)								((BP_SETTLUT(X) >> 10) & 0x003)
#define TLUT_TMEM_BASE(X)							(((BP_SETTLUT(X) >>  0) & 0x3ff) << 5)

#define TEX_MODE_MAX_LOD(X)						((BP_SETMODE1(X) >>  8) & 0x0ff)
#define TEX_MODE_MIN_LOD(X)						((BP_SETMODE1(X) >>  0) & 0x0ff)
#define TEX_MODE_LOD_CLAMP(X)					((BP_SETMODE0(X) >> 21) & 0x007)
#define TEX_MODE_MAX_ANISO(X)					((BP_SETMODE0(X) >> 19) & 0x003)
#define TEX_MODE_LOD_BIAS(X)					((BP_SETMODE0(X) >>  9) & 0x3ff)
#define TEX_MODE_MIN_FILTER(X)				((BP_SETMODE0(X) >>  5) & 0x007)
#define TEX_MODE_MAG_FILTER(X)				((BP_SETMODE0(X) >>  4) & 0x001)
#define TEX_MODE_WRAP_T(X)						((BP_SETMODE0(X) >>  2) & 0x003)
#define TEX_MODE_WRAP_S(X)						((BP_SETMODE0(X) >>  0) & 0x003)

#define SCISSORS_X										((BP_SCISSORS_XY >> 12) & 0xfff)
#define SCISSORS_Y										((BP_SCISSORS_XY >>  0) & 0xfff)
#define SCISSORS_W										((BP_SCISSORS_WH >> 12) & 0xfff)
#define SCISSORS_H										((BP_SCISSORS_WH >>  0) & 0xfff)

// direct formats
#define TEX_FORMAT_I4									0
#define TEX_FORMAT_I8									1
#define TEX_FORMAT_IA4								2
#define TEX_FORMAT_IA8								3
#define TEX_FORMAT_RGB565							4
#define TEX_FORMAT_RGB5A3							5
#define TEX_FORMAT_RGBA8							6
#define TEX_FORMAT_CMP								14

// indexed formats
#define TEX_FORMAT_CI4								8
#define TEX_FORMAT_CI8								9
#define TEX_FORMAT_CI14X2							10

#define TLUT_FORMAT_IA8								0
#define TLUT_FORMAT_RGB565						1
#define TLUT_FORMAT_RGB5A3						2

#define GL_TEXTURE_RECTANGLE					GL_TEXTURE_RECTANGLE_NV


// 0, DIRECT_U8, DIRECT_S8, ..., I8_U8, I8_S8, ..., I16_U8, I16_S8, ...
// tt c fff
#define GXIDX(X,v) ((VCD_##X == 0) ? 0 : \
 ((((VCD_##X - 1) << 4) | (VAT_##X##_CNT(v) << 3) | VAT_##X##_FMT(v)) + 1))

// VCD:
//   0 - no data
//   1 - direct
//   2 - 8 bit index
//   3 - 16 bit index

#define GPOP(X) 						(gpop[X])
#define GPOPR(m,n,X)\
	({\
		int i;\
		for (i = m; i <= n; i++)\
			GPOP (i) = X;\
	})


#define TMEM_SIZE						0x100000
#define TMEM_MASK						0x0fffff
#define TMEM_ADDRESS(X)			(&TMEM ((X) & TMEM_MASK))

#define BP(X)								(gxs.bp[X])
#define CP(X)								(gxs.cp[X])
#define XF(X)								(gxs.xf[(X) & 0xff])
#define FXF(X)							(((float *) gxs.xf)[(X) & 0xff])

#define XFMAT(X)						(gxs.xfmat[X])
#define XFMAT_TEX(X)				(gxs.xfmat[X + (0x0500 / 2)])
#define TMEM(X)							(gxs.tmem[X])

#define TEX_DISABLED					0
#define TEX_ENABLED_P2				1
#define TEX_ENABLED						2

#define GX_TOGGLE_WIREFRAME					0
#define GX_TOGGLE_TEX_RELOAD				1
#define GX_TOGGLE_TEX_DUMP					2
#define GX_TOGGLE_FORCE_LINEAR			3
#define GX_TOGGLE_TEX_TRANSPARENT		4
#define GX_TOGGLE_DRAW							5
#define GX_TOGGLE_NO_LOGIC_OPS			6
#define GX_TOGGLE_FULLBRIGHT				7

typedef struct
{
	__u32 cp[0x100];
	__u32 xf[0x100];
	__u32 bp[0x100];

	// matrix memory
	float xfmat[0x800];

	// texture memory (used for tluts)
	__u8 tmem[TMEM_SIZE];
} GXState;

typedef struct
{
	__u32 address;
	__u32 tlut_address;
	unsigned int tex;
	unsigned int type;

	unsigned int p2;
	int mipmap;
	int reload;
	
	int width, height;
	int format;
	int tlut_format;
} TextureTag;

typedef struct
{
	unsigned int format, type, internal_format;
} TexFormat;


extern GXState gxs;


void gx_init (void);
void gx_reinit (void);
int gx_parse_list (__u32 address, __u32 length);

char *gx_convert_texture (__u32 address, int width, int height, int format,
													__u32 tlut_address, int tlut_format);

int gx_switch (int sw);


#endif // __HW_GX_H
